# -*- coding: utf-8 -*-
"""Test suite for axonius_api_client.tools."""
import io
import os

import pytest
from click.testing import CliRunner

from ....cli import cli
from ....cli.grp_tools import cmd_shell
from ....tools import pathlib
from ...utils import MockError, load_clirunner


class TestCmdShell:
    """Pass."""

    def test_prompt(self, request, monkeypatch):
        """Pass."""
        runner = CliRunner(mix_stderr=False)

        url = request.config.getoption("--ax-url")
        key = request.config.getoption("--ax-key")
        secret = request.config.getoption("--ax-secret")

        monkeypatch.delenv("AX_URL", raising=False)
        monkeypatch.delenv("AX_KEY", raising=False)
        monkeypatch.delenv("AX_SECRET", raising=False)
        prompt_input = "\n".join([url, key, secret, "exit()"])

        args = ["tools", "shell"]

        with runner.isolated_filesystem():
            histpath = pathlib.Path(os.getcwd())
            histfile = histpath / cmd_shell.HISTFILE
            monkeypatch.setattr(cmd_shell, "HISTPATH", format(histpath))

            result1 = runner.invoke(cli=cli, args=args, input=prompt_input)

            assert histfile.is_file(), list(histpath.iterdir())

        exit_code1 = result1.exit_code
        stdout1 = result1.stdout
        stderr1 = result1.stderr

        assert stdout1
        assert stderr1
        assert exit_code1 == 0

        outlines1 = stdout1.splitlines()

        assert outlines1[0] == "URL: {}".format(url)
        assert outlines1[1] == "API Key of user: "
        assert outlines1[2] == "API Secret of user: "

    def test_no_prompt(self, request, monkeypatch):
        """Pass."""
        runner = load_clirunner(request, monkeypatch)

        prompt_input = "\n".join(["exit()"])

        args = ["tools", "shell"]

        with runner.isolated_filesystem():
            histpath = pathlib.Path(os.getcwd())
            histfile = histpath / cmd_shell.HISTFILE
            monkeypatch.setattr(cmd_shell, "HISTPATH", format(histpath))

            result1 = runner.invoke(cli=cli, args=args, input=prompt_input)

            assert histfile.is_file(), list(histpath.iterdir())
        exit_code1 = result1.exit_code
        stdout1 = result1.stdout
        stderr1 = result1.stderr

        assert stdout1
        assert stderr1
        assert exit_code1 == 0

        assert exit_code1 == 0


class TestCliWriteHistFile:
    """Pass."""

    def test_default(self, monkeypatch):
        """Pass."""
        runner = CliRunner(mix_stderr=False)

        with runner.isolated_filesystem():
            histpath = pathlib.Path(os.getcwd())
            histfile = histpath / cmd_shell.HISTFILE
            monkeypatch.setattr(cmd_shell, "HISTPATH", format(histpath))
            cmd_shell.write_hist_file()
            assert histfile.is_file(), list(histpath.iterdir())


class TestCliRegisterReadline:
    """Pass."""

    def test_default(self, monkeypatch):
        """Pass."""
        runner = CliRunner(mix_stderr=False)

        with runner.isolated_filesystem():
            histpath = pathlib.Path(os.getcwd())
            histfile = histpath / cmd_shell.HISTFILE
            monkeypatch.setattr(cmd_shell, "HISTPATH", format(histpath))

            cmd_shell.register_readline()
            assert histfile.is_file(), list(histpath.iterdir())

    def test_exc(self, monkeypatch, capsys):
        """Pass."""
        runner = CliRunner(mix_stderr=False)

        with runner.isolated_filesystem():
            monkeypatch.setattr(cmd_shell, "atexit", MockError)

            histpath = pathlib.Path(os.getcwd())
            histfile = histpath / cmd_shell.HISTFILE
            monkeypatch.setattr(cmd_shell, "HISTPATH", format(histpath))

            cmd_shell.register_readline()

            assert histfile.is_file(), list(histpath.iterdir())

        captured = capsys.readouterr()

        assert (
            captured.err.splitlines()[0]
            == "** ERROR: Unable to register history and autocomplete:"
        )


class TestCliJdump:
    """Pass."""

    def test_default(self, capsys):
        """Pass."""
        cmd_shell.jdump([])
        captured = capsys.readouterr()
        assert captured.out.splitlines()[0] == "[]"


class TestCliSpawnShell:
    """Pass."""

    def test_default(self, monkeypatch):
        """Pass."""
        runner = CliRunner(mix_stderr=False)

        monkeypatch.setattr("sys.stdin", io.StringIO("exit()"))

        with runner.isolated_filesystem():
            histpath = pathlib.Path(os.getcwd())
            monkeypatch.setattr(cmd_shell, "HISTPATH", format(histpath))

            with pytest.raises(SystemExit):
                cmd_shell.spawn_shell()



# -*- coding: utf-8 -*-
"""Command line interface for Axonius API Client."""
from ..context import CONTEXT_SETTINGS, click
from ..options import AUTH, QUERY, add_options, get_option_help

OPTIONS = [
    *AUTH,
    *QUERY,
    get_option_help(choices=["auth", "query"]),
]


@click.command(name="count", context_settings=CONTEXT_SETTINGS)
@add_options(OPTIONS)
@click.pass_context
def cmd(ctx, url, key, secret, query_file, help_detailed=None, **kwargs):
    """Get the count of assets from a query."""
    if query_file:
        kwargs["query"] = query_file.read().strip()

    client = ctx.obj.start_client(url=url, key=key, secret=secret)

    p_grp = ctx.parent.command.name
    apiobj = getattr(client, p_grp)

    with ctx.obj.exc_wrap(wraperror=ctx.obj.wraperror):
        raw_data = apiobj.count(**kwargs)

    click.secho(format(raw_data))



# -*- coding: utf-8 -*-
"""Command line interface for Axonius API Client."""
from ....tools import listify
from ...context import CONTEXT_SETTINGS, click
from ...options import AUTH, add_options

OPTIONS = [*AUTH]


@click.command(name="get-tags", context_settings=CONTEXT_SETTINGS)
@add_options(OPTIONS)
@click.pass_context
def cmd(ctx, url, key, secret, **kwargs):
    """Get all known tags for saved queries."""
    client = ctx.obj.start_client(url=url, key=key, secret=secret)

    p_grp = ctx.parent.parent.command.name
    apiobj = getattr(client, p_grp)

    with ctx.obj.exc_wrap(wraperror=ctx.obj.wraperror):
        rows = listify(apiobj.saved_query.get_tags(**kwargs))

    ctx.obj.echo_ok(f"Successfully fetched {len(rows)} saved query tags")
    click.secho("\n".join(rows))
    ctx.exit(0)



# -*- coding: utf-8 -*-
"""Test suite."""
import copy
import warnings

import pytest

from axonius_api_client.constants import CSV_ADAPTER, DEFAULT_NODE
from axonius_api_client.exceptions import (
    ApiError,
    ConfigUnchanged,
    ConfigUnknown,
    NotFoundError,
)

from ...meta import (
    CSV_FILECONTENT_BYTES,
    CSV_FILECONTENT_STR,
    CSV_FILENAME,
    FIELD_FORMATS,
    SCHEMA_TYPES,
)


def val_parsed_schema(schema):
    """Pass."""
    for setting_name, item in schema.items():
        item_name = item.pop("name")
        assert isinstance(item_name, str) and item_name
        assert setting_name == item_name

        item_type = item.pop("type")
        assert isinstance(item_type, str) and item_type
        assert item_type in SCHEMA_TYPES

        item_required = item.pop("required")
        assert isinstance(item_required, bool)

        item_title = item.pop("title")
        assert isinstance(item_title, str) and item_title

        # optionals
        item_format = item.pop("format", "")
        assert isinstance(item_format, str)
        assert item_format in FIELD_FORMATS or item_format == ""

        item_description = item.pop("description", "")
        assert isinstance(item_description, str)

        item_enum = item.pop("enum", [])
        assert isinstance(item_enum, list)
        for x in item_enum:
            assert isinstance(x, str) and x

        item_default = item.pop("default", "")
        assert isinstance(item_default, (str, int, bool)) or item_default in [
            None,
            [],
        ]

        item_items = item.pop("items", {})
        assert isinstance(item_items, dict)

        item_report_id = item.pop("report_id", "")
        assert isinstance(item_report_id, str)

        assert not item


class TestAdaptersBase:
    """Pass."""

    @pytest.fixture(scope="class")
    def apiobj(self, api_adapters):
        """Pass."""
        return api_adapters

    @pytest.fixture(scope="class")
    def adapter(self, apiobj):
        """Pass."""
        return apiobj.get_by_name(name=CSV_ADAPTER, node=DEFAULT_NODE)


class TestAdaptersPrivate(TestAdaptersBase):
    """Pass."""

    def test_private_get(self, apiobj):
        """Pass."""
        adapters = apiobj._get()
        assert isinstance(adapters, dict)
        self.val_raw_adapters(adapters=adapters)

    def val_raw_adapters(self, adapters):
        """Pass."""
        for name, instances in adapters.items():
            assert name.endswith("_adapter")
            assert isinstance(instances, list)

            for instance in instances:
                self.val_raw_adapter(name=name, instance=instance)

    def val_raw_adapter(self, name, instance):
        """Pass."""
        node_id = instance.pop("node_id")
        assert isinstance(node_id, str) and node_id

        unique_plugin_name = instance.pop("unique_plugin_name")
        assert isinstance(unique_plugin_name, str) and unique_plugin_name

        node_name = instance.pop("node_name")
        assert isinstance(node_name, str) and node_name

        clients = instance.pop("clients")
        assert isinstance(clients, list)

        config = instance.pop("config")
        assert isinstance(config, dict) and config
        assert len(config) in [1, 2]
        assert "AdapterBase" in config

        for config_name, item in config.items():
            item_config = item.pop("config")
            item_schema = item.pop("schema")
            item_pretty_name = item_schema.pop("pretty_name")

            assert isinstance(item_config, dict) and item_config
            assert isinstance(item_pretty_name, str) and item_pretty_name
            assert isinstance(item_schema, dict) and item_schema

            self.val_raw_schema(name=name, schema=item_schema)

            assert not item

        schema = instance.pop("schema")
        assert isinstance(schema, dict) and schema
        self.val_raw_schema(name=name, schema=schema)

        status = instance.pop("status")
        assert status in ["warning", "success", None, ""]

        supported_features = instance.pop("supported_features")
        assert isinstance(supported_features, list)
        for x in supported_features:
            assert isinstance(x, str)

        assert not instance

        for client in clients:
            self.val_raw_client(name=name, client=client, instance_status=status)

    def val_raw_client(self, name, client, instance_status):
        """Pass."""
        assert isinstance(client, dict)

        client_config = client.pop("client_config")
        assert isinstance(client_config, dict) and client_config

        uuid = client.pop("uuid")
        assert isinstance(uuid, str) and uuid

        client_id = client.pop("client_id")
        assert isinstance(client_id, str)

        if not client_id:
            msg = "Client for {} has an empty client_id {}"
            msg = msg.format(name, client_id)
            warnings.warn(msg)

        error = client.pop("error")
        assert isinstance(error, str) or error is None

        node_id = client.pop("node_id")
        assert isinstance(node_id, str) and node_id

        status = client.pop("status")
        assert isinstance(status, str)
        assert status in ["warning", "error", "success"]

        if status == "error":
            assert instance_status == "warning"

        date_fetched = client.pop("date_fetched")
        assert isinstance(date_fetched, str)

        adapter_name = client.pop("adapter_name")
        assert isinstance(adapter_name, str)

        assert not client

    def val_raw_schema(self, name, schema):
        """Pass."""
        assert isinstance(schema, dict)

        items = schema.pop("items")
        assert isinstance(items, list) and items

        item_names = [x["name"] for x in items]

        for item in items:
            assert isinstance(item, dict)

            item_name = item.pop("name")
            assert isinstance(item_name, str) and item_name

            item_type = item.pop("type")
            assert isinstance(item_type, str) and item_type

            item_title = item.pop("title")
            assert isinstance(item_title, str) and item_title

            item_format = item.pop("format", "")
            assert isinstance(item_format, str)

            item_description = item.pop("description", "")
            assert isinstance(item_description, str)

            item_enum = item.pop("enum", [])
            assert isinstance(item_enum, list)
            for x in item_enum:
                assert isinstance(x, str)

            item_default = item.pop("default", "")
            assert isinstance(item_default, (str, int, bool)) or item_default is None

            item_items = item.pop("items", {})
            assert isinstance(item_items, dict)

            item_req = item.pop("required", False)
            assert isinstance(item_req, bool)
            if item_type not in SCHEMA_TYPES:
                msg = "Setting for adapter {!r} has an unexpected item_type {!r}"
                msg = msg.format(name, item_type)
                warnings.warn(msg)

            item_report_id = item.pop("report_id", "")
            assert isinstance(item_report_id, str)

            assert not item

        required = schema.pop("required")
        assert isinstance(required, list)
        for req in required:
            assert isinstance(req, str)

            if req not in item_names:
                msg = "Schema for {} has required item {!r} not in defined items {}"
                msg = msg.format(name, req, item_names)
                warnings.warn(msg)

        schema_type = schema.pop("type")
        assert schema_type == "array"

        assert not schema

    def test_private_file_upload_str(self, apiobj, adapter):
        """Pass."""
        data = apiobj._file_upload(
            name_raw=adapter["name_raw"],
            node_id=adapter["node_id"],
            field_name=CSV_ADAPTER,
            file_name=CSV_FILENAME,
            file_content=CSV_FILECONTENT_STR,
        )
        assert isinstance(data, dict)
        assert data["uuid"]
        assert data["filename"]

    def test_private_file_upload_bytes(self, apiobj, adapter):
        """Pass."""
        data = apiobj._file_upload(
            name_raw=adapter["name_raw"],
            node_id=adapter["node_id"],
            field_name=CSV_ADAPTER,
            file_name=CSV_FILENAME,
            file_content=CSV_FILECONTENT_BYTES,
        )
        assert isinstance(data, dict)
        assert data["uuid"]
        assert data["filename"]

    def test_private_config_update(self, apiobj, adapter):
        """Pass."""
        current = apiobj._config_get(
            name_plugin=adapter["name_plugin"], name_config="AdapterBase"
        )
        key = "user_last_fetched_threshold_hours"
        current_config = current["config"]
        config_update = copy.deepcopy(current_config)
        set_value = config_update[key] + 1
        config_update[key] = set_value

        set_response = apiobj._config_update(
            name_raw=adapter["name_raw"],
            name_config="AdapterBase",
            new_config=config_update,
        )

        assert not set_response
        updated = apiobj._config_get(
            name_plugin=adapter["name_plugin"], name_config="AdapterBase"
        )
        updated_config = updated["config"]
        assert updated_config[key] == set_value
        assert updated_config == config_update
        assert updated_config != current_config

        reconfig_update = copy.deepcopy(updated_config)
        reconfig_update[key] = current_config[key]
        reset_response = apiobj._config_update(
            name_raw=adapter["name_raw"],
            name_config="AdapterBase",
            new_config=reconfig_update,
        )
        assert not reset_response
        post_reset = apiobj._config_get(
            name_plugin=adapter["name_plugin"], name_config="AdapterBase"
        )
        assert post_reset["config"] == current_config

    def test_private_config_get_generic(self, apiobj):
        """Pass."""
        for meta in apiobj.get():
            data = apiobj._config_get(
                name_plugin=meta["name_plugin"],
                name_config=meta["schemas"]["generic_name"],
            )
            assert isinstance(data, dict)

            config = data.pop("config")
            assert isinstance(config, dict)

            schema = data.pop("schema")
            assert isinstance(schema, dict)
            # XXX test schema!

            assert not data

            items = schema.pop("items")
            assert isinstance(items, list)

            pretty_name = schema.pop("pretty_name")
            assert isinstance(pretty_name, str) and pretty_name

            required = schema.pop("required")
            assert isinstance(required, list)
            assert all([isinstance(x, str) for x in required])

            stype = schema.pop("type")
            assert stype == "array"

            assert not schema

    def test_private_config_get_specific(self, apiobj):
        """Pass."""
        for meta in apiobj.get():
            if not meta["schemas"]["specific_name"]:
                continue

            data = apiobj._config_get(
                name_plugin=meta["name_plugin"],
                name_config=meta["schemas"]["specific_name"],
            )
            assert isinstance(data, dict)

            config = data.pop("config")
            assert isinstance(config, dict)

            schema = data.pop("schema")
            assert isinstance(schema, dict)

            assert not data

            items = schema.pop("items")
            assert isinstance(items, list)

            pretty_name = schema.pop("pretty_name")
            assert isinstance(pretty_name, str) and pretty_name

            required = schema.pop("required")
            assert isinstance(required, list)
            assert all([isinstance(x, str) for x in required])

            stype = schema.pop("type")
            assert stype == "array"

            assert not schema


class TestAdaptersPublic(TestAdaptersBase):
    """Pass."""

    def test_get(self, apiobj):
        """Pass."""
        adapters = apiobj.get()
        assert all(["schemas" in x for x in adapters])
        self.val_parsed_adapters(adapters=adapters)

    def val_parsed_adapters(self, adapters):
        """Pass."""
        assert isinstance(adapters, list)
        for adapter in adapters:
            self.val_parsed_adapter(adapter=adapter)

    def val_parsed_adapter(self, adapter):
        """Pass."""
        assert isinstance(adapter, dict)

        name = adapter.pop("name")
        assert isinstance(name, str)

        name_raw = adapter.pop("name_raw")
        assert isinstance(name_raw, str)

        name_plugin = adapter.pop("name_plugin")
        assert isinstance(name_plugin, str)

        node_name = adapter.pop("node_name")
        assert isinstance(node_name, str)

        node_id = adapter.pop("node_id")
        assert isinstance(node_id, str)

        status = adapter.pop("status")
        assert isinstance(status, str)
        assert status in ["warning", "success", ""]

        features = adapter.pop("features")
        assert isinstance(features, list)
        for x in features:
            assert isinstance(x, str) and x

        schemas = adapter.pop("schemas")
        assert isinstance(schemas, dict) and schemas

        schema_cnx = schemas.pop("cnx")
        assert isinstance(schema_cnx, dict) and schema_cnx

        generic_name = schemas.pop("generic_name")
        assert isinstance(generic_name, str) and generic_name
        assert generic_name == "AdapterBase"

        generic_schema = schemas.pop("generic")
        assert isinstance(generic_schema, dict) and generic_schema
        val_parsed_schema(schema=generic_schema)

        specific_name = schemas.pop("specific_name")
        assert isinstance(specific_name, str)

        specific_schema = schemas.pop("specific")
        assert isinstance(specific_schema, dict)

        assert not schemas

        if specific_name:
            assert specific_schema
            val_parsed_schema(schema=specific_schema)

        cnx_count_total = adapter.pop("cnx_count_total")
        assert isinstance(cnx_count_total, int)
        if not cnx_count_total:
            assert not status

        cnx_count_broken = adapter.pop("cnx_count_broken")
        assert isinstance(cnx_count_broken, int)

        cnx_count_working = adapter.pop("cnx_count_working")
        assert isinstance(cnx_count_working, int)
        if not cnx_count_broken and cnx_count_working:
            assert status == "success"

        cnxs = adapter.pop("cnx")
        assert isinstance(cnxs, list)
        for cnx in cnxs:
            self.val_parsed_cnx(
                cnx=cnx,
                adapter_name=name,
                adapter_name_raw=name_raw,
                adapter_node_id=node_id,
                adapter_node_name=node_name,
            )

        config = adapter.pop("config")
        assert isinstance(config, dict) and config

        generic_config = config.pop("generic")
        assert isinstance(generic_config, dict) and generic_config

        specific_config = config.pop("specific")
        assert isinstance(specific_config, dict)

        assert not adapter

    def val_parsed_cnx(
        self, cnx, adapter_name, adapter_name_raw, adapter_node_id, adapter_node_name,
    ):
        """Pass."""
        assert isinstance(cnx, dict)

        cnx_adapter_name = cnx.pop("adapter_name")
        assert cnx_adapter_name == adapter_name

        cnx_adapter_name_raw = cnx.pop("adapter_name_raw")
        assert cnx_adapter_name_raw == adapter_name_raw

        config = cnx.pop("config")
        assert isinstance(config, dict) and config

        date_fetched = cnx.pop("date_fetched")
        assert isinstance(date_fetched, str)

        error = cnx.pop("error")
        assert isinstance(error, str) or error is None

        cnx_id = cnx.pop("id")
        assert isinstance(cnx_id, str)

        cnx_node_id = cnx.pop("node_id")
        assert cnx_node_id == adapter_node_id

        cnx_node_name = cnx.pop("node_name")
        assert cnx_node_name == adapter_node_name

        working = cnx.pop("working")
        assert isinstance(working, bool)

        status = cnx.pop("status")
        assert isinstance(status, str) and status
        assert status in ["warning", "error", "success"]

        uuid = cnx.pop("uuid")
        assert isinstance(uuid, str) and uuid

        assert not cnx

    def test_get_by_name(self, apiobj):
        """Pass."""
        adapter = apiobj.get_by_name(name=CSV_ADAPTER, node=DEFAULT_NODE)
        assert "schemas" in adapter
        self.val_parsed_adapter(adapter=adapter)

    def test_get_by_name_bad_node(self, apiobj):
        """Pass."""
        with pytest.raises(NotFoundError):
            apiobj.get_by_name(name=CSV_ADAPTER, node="badwolf")

    def test_get_by_name_bad_name(self, apiobj):
        """Pass."""
        with pytest.raises(NotFoundError):
            apiobj.get_by_name(name="badwolf", node=DEFAULT_NODE)

    def test_config_get_generic_false_bad(self, apiobj):
        """Pass."""
        with pytest.raises(ApiError):
            apiobj.config_get(name=CSV_ADAPTER, node=DEFAULT_NODE, generic=False)

    def test_config_get_generic_false(self, apiobj):
        """Pass."""
        data = apiobj.config_get(name="aws", node=DEFAULT_NODE, generic=False)
        assert isinstance(data, dict)
        config = data.pop("config")
        assert isinstance(config, dict) and config

        schema = data.pop("schema")
        assert isinstance(schema, dict) and schema

        val_parsed_schema(schema=schema)

    def test_config_get_generic_true(self, apiobj):
        """Pass."""
        data = apiobj.config_get(name=CSV_ADAPTER, node=DEFAULT_NODE, generic=True)
        assert isinstance(data, dict)

        config = data.pop("config")
        assert isinstance(config, dict) and config

        schema = data.pop("schema")
        assert isinstance(schema, dict) and schema

        val_parsed_schema(schema=schema)

    def test_file_upload_str(self, apiobj, adapter):
        """Pass."""
        data = apiobj.file_upload(
            name=adapter["name"],
            node=adapter["node_name"],
            field_name=CSV_ADAPTER,
            file_name=CSV_FILENAME,
            file_content=CSV_FILECONTENT_STR,
        )
        assert isinstance(data, dict)
        assert data["uuid"]
        assert data["filename"]

    def test_file_upload_path(self, apiobj, tmp_path):
        """Pass."""
        test_path = tmp_path / CSV_FILENAME
        test_path.write_text(CSV_FILECONTENT_STR)

        data = apiobj.file_upload_path(
            name=CSV_ADAPTER, node=DEFAULT_NODE, path=test_path,
        )
        assert isinstance(data, dict)
        assert data["uuid"]
        assert data["filename"]

    def test_config_update_unknown(self, apiobj):
        """Pass."""
        with pytest.raises(ConfigUnknown):
            apiobj.config_update(
                name=CSV_ADAPTER, node=DEFAULT_NODE, generic=True, badwolf="badwolf",
            )

    def test_config_update_unchanged(self, apiobj):
        """Pass."""
        with pytest.raises(ConfigUnchanged):
            apiobj.config_update(name=CSV_ADAPTER, node=DEFAULT_NODE, generic=True)

    def test_config_update_generic_true(self, apiobj, adapter):
        """Pass."""
        data = apiobj.config_refetch(adapter=adapter)
        current = data["config"]
        key = "user_last_fetched_threshold_hours"
        current_value = current[key]
        new_value = current_value + 1

        set_response = apiobj.config_update(
            name=adapter["name"],
            node=adapter["node_name"],
            user_last_fetched_threshold_hours=new_value,
            generic=True,
        )
        set_value = set_response["config"][key]
        assert set_value == new_value

        reset_response = apiobj.config_update(
            name=adapter["name"],
            node=adapter["node_name"],
            user_last_fetched_threshold_hours=current_value,
            generic=True,
        )
        reset_value = reset_response["config"][key]
        assert reset_value == current_value



# -*- coding: utf-8 -*-
"""API models for working with device and user assets."""
import re

from ...constants import (AGG_ADAPTER_ALTS, AGG_ADAPTER_NAME, GET_SCHEMA_KEYS,
                          GET_SCHEMAS_KEYS)
from ...exceptions import ApiError, NotFoundError
from ...tools import listify, split_str, strip_right
from ..mixins import ChildMixins
from ..parsers.fields import parse_fields


class Fields(ChildMixins):
    """Child API model for working with fields for the parent asset type."""

    def get(self):
        """Get the schema of all adapters and their fields.

        Returns:
            :obj:`dict`: parsed output from :meth:`ParserFields.parse`
        """
        return parse_fields(raw=self._get())

    def get_adapter_names(self, value, fields_map=None):
        """Find an adapter by name regex."""
        fields_map = fields_map or self.get()

        search = strip_right(obj=value.lower().strip(), fix="_adapter")

        if search in AGG_ADAPTER_ALTS:
            search = AGG_ADAPTER_NAME

        search = re.compile(search, re.I)
        matches = [x for x in fields_map if search.search(x)]

        if not matches:
            msg = (
                "No adapter found where name regex matches {!r}, valid adapters:\n  {}"
            ).format(value, "\n  ".join(list(fields_map)))
            raise NotFoundError(msg)
        return matches

    def get_adapter_name(self, value, fields_map=None):
        """Find an adapter by name."""
        fields_map = fields_map or self.get()

        search = strip_right(obj=value.lower().strip(), fix="_adapter")

        if search in AGG_ADAPTER_ALTS:
            search = AGG_ADAPTER_NAME

        if search in fields_map:
            return search

        msg = "No adapter found where name equals {!r}, valid adapters:\n  {}"
        msg = msg.format(value, "\n  ".join(list(fields_map)))
        raise NotFoundError(msg)

    def get_field_schemas(self, value, schemas, **kwargs):
        """Find a schema for a field by regex of name."""
        keys = kwargs.get("keys", GET_SCHEMAS_KEYS)
        search = re.compile(value.lower().strip(), re.I)

        matches = []

        for schema in schemas:
            for key in keys:
                if search.search(schema[key]) and schema not in matches:
                    matches.append(schema)

        # XXX fix test case for this
        # os\. will fail for adapters that do not have os.type/dist/etc
        # if not matches:
        #     msg = "No field found where {} matches regex {!r}, valid fields: \n{}"
        #     msg = msg.format(
        #         keys, value, "\n".join(self._prettify_schemas(schemas=schemas)),
        #     )
        #     raise NotFoundError(msg)
        return matches

    def get_field_schema(self, value, schemas, **kwargs):
        """Find a schema for a field by name."""
        keys = kwargs.get("keys", GET_SCHEMA_KEYS)
        search = value.lower().strip()

        for schema in schemas:
            for key in keys:
                if search == schema[key]:
                    return schema

        msg = "No field found where any of {} equals {!r}, valid fields: \n{}"
        msg = msg.format(keys, value, "\n".join(self._prettify_schemas(schemas=schemas)))
        raise NotFoundError(msg)

    def get_field_name(self, value, field_manual=False, fields_map=None):
        """Pass."""
        if field_manual:
            return value

        adapter, fields = self.split_search(value=value)

        if len(fields) != 1:
            raise ApiError("More than one field supplied to {}".format(value))

        field = fields[0]

        fields_map = fields_map or self.get()
        adapter = self.get_adapter_name(value=adapter, fields_map=fields_map)
        schemas = fields_map[adapter]
        schema = self.get_field_schema(value=field, schemas=schemas)
        return schema["name_qual"]

    def get_field_names_re(self, value, fields_map=None):
        """Pass."""
        splits = self.split_searches(value=value)
        fields_map = fields_map or self.get()

        matches = []

        for adapter_re, fields in splits:
            adapters = self.get_adapter_names(value=adapter_re, fields_map=fields_map)

            for adapter in adapters:
                for field in fields:
                    fschemas = self.get_field_schemas(
                        value=field, schemas=fields_map[adapter]
                    )
                    names = [x["name_qual"] for x in fschemas]
                    matches += [x for x in names if x not in matches]
        return matches

    def get_field_names_eq(self, value, fields_map=None):
        """Pass."""
        splits = self.split_searches(value=value)
        fields_map = fields_map or self.get()

        matches = []

        for adapter_name, names in splits:
            adapter = self.get_adapter_name(value=adapter_name, fields_map=fields_map)
            for name in names:
                schemas = fields_map[adapter]
                schema = self.get_field_schema(value=name, schemas=schemas)
                if schema["name_qual"] not in matches:
                    matches.append(schema["name_qual"])

        return matches

    def validate(
        self,
        fields=None,
        fields_regex=None,
        fields_manual=None,
        fields_default=True,
        fields_map=None,
    ):
        """Validate provided fields."""
        fields_manual = listify(obj=fields_manual)

        selected = []

        if fields_default:
            selected += self.parent.fields_default

        if fields_manual:
            selected += [x for x in fields_manual if x not in selected]

        if not any([fields, fields_regex]):
            if not selected:
                raise ApiError("No fields supplied, must supply at least one field")
            return selected

        fields_map = fields_map or self.get()

        for field in self.get_field_names_eq(value=fields, fields_map=fields_map):
            if field not in selected:
                selected.append(field)

        for field in self.get_field_names_re(value=fields_regex, fields_map=fields_map):
            if field not in selected:
                selected.append(field)

        return selected

    def split_searches(self, value):
        """Pass."""
        return [self.split_search(value=x) for x in listify(obj=value)]

    def split_search(self, value, adapter=AGG_ADAPTER_NAME):
        """Pass."""
        search = value.strip().lower()

        if ":" in search:
            adapter_split, field = search.split(":", 1)
            if not adapter_split:
                adapter_split = adapter
        else:
            field = search
            adapter_split = adapter

        adapter_split = strip_right(obj=adapter_split.lower().strip(), fix="_adapter")

        fields = split_str(
            obj=field, split=",", strip=None, do_strip=True, lower=True, empty=False,
        )

        if not fields:
            msg = "No fields provided in {!r}, format must be 'adapter:field'"
            msg = msg.format(value)
            raise ApiError(msg)

        return adapter_split, fields

    def _get(self):
        """Direct API method to get the schema of all fields.

        Returns:
            :obj:`dict`: schema of all fields
        """
        return self.request(method="get", path=self.router.fields)

    def _prettify_schemas(self, schemas):
        """Pass."""
        stmpl = "{adapter_name}:{name_base:{name_base_len}} -> {column_title}".format
        name_base_len = max([len(x["name_base"]) for x in schemas])
        return [stmpl(name_base_len=name_base_len, **x) for x in schemas]



# -*- coding: utf-8 -*-
"""Command line interface for Axonius API Client."""
from ....tools import json_dump
from ...context import CONTEXT_SETTINGS, click
from ...options import AUTH, add_options
from .grp_common import EXPORT, str_sections

OPTIONS = [*AUTH, EXPORT]


@click.command(name="get", context_settings=CONTEXT_SETTINGS)
@add_options(OPTIONS)
@click.pass_context
def cmd(ctx, url, key, secret, export_format, **kwargs):
    """Get all settings."""
    client = ctx.obj.start_client(url=url, key=key, secret=secret)

    apiname = ctx.parent.command.name.replace("-", "_")
    apiobj = getattr(client.system, apiname)

    with ctx.obj.exc_wrap(wraperror=ctx.obj.wraperror):
        settings = apiobj.get()

    if export_format == "str":
        str_sections(meta=settings)
        ctx.exit(0)

    if export_format == "json-config":
        config = settings["config"]
        click.secho(json_dump(config))
        ctx.exit(0)

    if export_format == "json-full":
        click.secho(json_dump(settings))
        ctx.exit(0)

    ctx.exit(1)



# -*- coding: utf-8 -*-
"""API models for working with device and user assets."""

import tabulate

from ...constants import TABLE_FORMAT, TABLE_MAX_ROWS
from ...exceptions import ApiError
from ...tools import listify
from .base import Base


class Table(Base):
    """Pass."""

    CB_NAME = "table"

    def start(self, **kwargs):
        """Create file descriptor."""
        super(Table, self).start(**kwargs)
        self.GETARGS["field_null"] = True
        self.GETARGS["field_flatten"] = True
        self.GETARGS["field_join"] = True
        self.GETARGS["field_titles"] = True

        self._table_format = self.GETARGS.get("table_format", TABLE_FORMAT)
        self.check_fmt(fmt=self._table_format)
        self._rows = []
        self.open_fd()

    def stop(self, **kwargs):
        """Close file descriptor."""
        super(Table, self).stop(**kwargs)
        rows = getattr(self, "_rows", [])

        table = tabulate.tabulate(
            tabular_data=rows,
            tablefmt=self._table_format,
            showindex=False,
            headers="keys",
        )

        self._fd.write(table)
        self._fd.write("\n")
        self.close_fd()

    def row(self, row):
        """Process row."""
        self.check_stop()
        rows = super(Table, self).row(row=row)
        # XXX textwrap key/values
        self._rows += listify(rows)
        return rows

    def check_stop(self):
        """Pass."""
        max_rows = self.GETARGS.get("table_max_rows", TABLE_MAX_ROWS)
        rows_processed = self.STATE.get("rows_processed", 0)

        if all([rows_processed, max_rows]) and rows_processed >= max_rows:
            self.STATE["stop"] = True
            self.STATE["stop_msg"] = f"table_max_rows of {max_rows}"

    def check_fmt(self, fmt):
        """Pass."""
        if fmt not in tabulate.tabulate_formats:
            fmts = ", ".join(tabulate.tabulate_formats)
            msg = f"{fmt!r} is not a valid table format, must be one of {fmts}"
            self.echo(msg=msg, error=ApiError)

    @property
    def args_map(self):
        """Pass."""
        args = super(Table, self).args_map
        return args + [
            ["table_format", "Use table format", TABLE_FORMAT],
            ["table_max_rows", "Maximum table rows", TABLE_MAX_ROWS],
        ]



# -*- coding: utf-8 -*-
"""Logging utilities."""
import logging
import logging.handlers
import time

from . import __package__ as PACKAGE_ROOT
from .constants import (
    LOG_DATEFMT_CONSOLE,
    LOG_DATEFMT_FILE,
    LOG_FILE_MAX_FILES,
    LOG_FILE_MAX_MB,
    LOG_FILE_NAME,
    LOG_FILE_PATH,
    LOG_FILE_PATH_MODE,
    LOG_FMT_CONSOLE,
    LOG_FMT_FILE,
    LOG_LEVEL_CONSOLE,
    LOG_LEVEL_FILE,
    LOG_LEVELS_INT_CSV,
    LOG_LEVELS_STR_CSV,
    LOG_NAME_FILE,
    LOG_NAME_STDERR,
    LOG_NAME_STDOUT,
)
from .exceptions import ToolsError
from .tools import get_path, is_int


def gmtime():
    """Set the logging system to use GMT for time strings."""
    logging.Formatter.converter = time.gmtime


def localtime():
    """Set the logging system to use local time for time strings."""
    logging.Formatter.converter = time.localtime


def get_obj_log(obj, level=None, **kwargs):
    """Get a child logger for an object.

    Notes:
        * level is resolved using resolved using :meth:`str_level`

    Args:
        obj (:obj:`object`): object to get a logger for
        level (:obj:`str` or :obj:`int`, optional): default ``None`` - level to set
        **kwargs:
            * logger (:obj:`logging.Logger`): default ``module of class of obj`` -
              logger to get child from

    Returns:
        :obj:`logging.Logger`: created logger child obj
    """
    logger = kwargs.get("logger", logging.getLogger(obj.__class__.__module__))
    log = logger.getChild(obj.__class__.__name__)
    set_log_level(obj=log, level=level)
    return log


def set_log_level(obj, level=None):
    """Set a logger or handler to a log level.

    Notes:
        * level is resolved using resolved using :meth:`str_level`

    Args:
        obj (:obj:`logging.Logger` or :obj:`logging.Handler`): object to set lvl on
        level (:obj:`str` or :obj:`int`): default ``None`` - level to set
    """
    if level:
        obj.setLevel(getattr(logging, str_level(level=level)))


def str_level(level):
    """Get a logging level in str format.

    Args:
        level (:obj:`str` or :obj:`int`): level to get str format of

    Returns:
        :obj:`str`: str repr of logging level

    Raises:
        :exc:`ToolsError`: if level is not mappable as an int or str
            to known logger level in :mod:`logging`
    """
    if is_int(obj=level, digit=True):
        level_mapped = logging.getLevelName(int(level))
        if hasattr(logging, level_mapped):
            return level_mapped

    if isinstance(level, str):
        if hasattr(logging, level.upper()):
            return level.upper()

    error = (
        f"Invalid logging level {level!r}, must be one of "
        f"{LOG_LEVELS_STR_CSV} or {LOG_LEVELS_INT_CSV}"
    )
    raise ToolsError(error)


def add_stderr(obj, **kwargs):
    """Add a StreamHandler to a logger object that outputs to STDERR.

    Args:
        obj (:obj:`logging.Logger`): logger obj to add handler to
        **kwargs:
            * level (:obj:`logging.Logger`):
              default: :data:`axonius_api_client.LOG_LEVEL_CONSOLE` -
              log level to assign to handler
            * hname (:obj:`str`):
              default: :data:`axonius_api_client.LOG_NAME_STDERR` -
              name to assign to handler
            * fmt (:obj:`str`):
              default: :data:`axonius_api_client.LOG_FMT_CONSOLE` -
              logging format to use
            * datefmt (:obj:`str`):
              default: :data:`axonius_api_client.LOG_DATEFMT_CONSOLE` -
              date format to use

    Returns:
        :obj:`logging.StreamHandler`: handler that was added to logger obj
    """
    level = kwargs.get("level", LOG_LEVEL_CONSOLE)
    hname = kwargs.get("hname", LOG_NAME_STDERR)
    fmt = kwargs.get("fmt", LOG_FMT_CONSOLE)
    datefmt = kwargs.get("datefmt", LOG_DATEFMT_CONSOLE)
    htype = logging.StreamHandler

    return add_handler(
        obj=obj, hname=hname, htype=htype, level=level, fmt=fmt, datefmt=datefmt
    )


def add_stdout(obj, **kwargs):
    """Add a StreamHandler to a logger object that outputs to STDOUT.

    Args:
        obj (:obj:`logging.Logger`): logger obj to add handler to
        **kwargs:
            * level (:obj:`logging.Logger`):
              default: :data:`axonius_api_client.LOG_LEVEL_CONSOLE` -
              log level to assign to handler
            * hname (:obj:`str`):
              default: :data:`axonius_api_client.LOG_NAME_STDOUT` -
              name to assign to handler
            * fmt (:obj:`str`):
              default: :data:`axonius_api_client.LOG_FMT_CONSOLE` -
              logging format to use
            * datefmt (:obj:`str`):
              default: :data:`axonius_api_client.LOG_DATEFMT_CONSOLE` -
              date format to use

    Returns:
        :obj:`logging.StreamHandler`: handler that was added to logger obj
    """
    level = kwargs.get("level", LOG_LEVEL_CONSOLE)
    hname = kwargs.get("hname", LOG_NAME_STDOUT)
    fmt = kwargs.get("fmt", LOG_FMT_CONSOLE)
    datefmt = kwargs.get("datefmt", LOG_DATEFMT_CONSOLE)
    htype = logging.StreamHandler

    return add_handler(
        obj=obj, hname=hname, htype=htype, level=level, fmt=fmt, datefmt=datefmt
    )


def add_file(obj, **kwargs):
    """Add a RotatingFileHandler to a logger object.

    Args:
        obj (:obj:`logging.Logger`): logger obj to add handler to
        **kwargs:
            * level (:obj:`logging.Logger`):
              default: :data:`axonius_api_client.LOG_LEVEL_FILE` -
              log level to assign to handler
            * hname (:obj:`str`):
              default: :data:`axonius_api_client.LOG_NAME_FILE` -
              name to assign to handler
            * fmt (:obj:`str`):
              default: :data:`axonius_api_client.LOG_FMT_FILE` -
              logging format to use
            * datefmt (:obj:`str`):
              default: :data:`axonius_api_client.LOG_DATEFMT_FILE` -
              date format to use
            * file_path (:obj:`str`):
              default: :data:`axonius_api_client.LOG_FILE_PATH` -
              path to write file_name to
            * file_name (:obj:`str`):
              default: :data:`axonius_api_client.LOG_FILE_NAME` -
              name of file to write log entries to
            * file_path_mode (:obj:`oct`):
              default: :data:`axonius_api_client.LOG_FILE_PATH_MODE` -
              permissions to assign to directory for log file if it has to be
              created
            * max_mb (:obj:`int`):
              default: :data:`axonius_api_client.LOG_FILE_MAX_MB` -
              rollover trigger in MB
            * max_files (:obj:`int`):
              default: :data:`axonius_api_client.LOG_FILE_MAX_FILES` -
              max files to keep for rollover

    Returns:
        :obj:`logging.StreamHandler`: handler that was added to logger obj
    """
    level = kwargs.get("level", LOG_LEVEL_FILE)
    hname = kwargs.get("hname", LOG_NAME_FILE)
    file_path = kwargs.get("file_path", LOG_FILE_PATH)
    file_name = kwargs.get("file_name", LOG_FILE_NAME)
    file_path_mode = kwargs.get("file_path_mode", LOG_FILE_PATH_MODE)
    max_mb = kwargs.get("max_mb", LOG_FILE_MAX_MB)
    max_files = kwargs.get("max_files", LOG_FILE_MAX_FILES)
    fmt = kwargs.get("fmt", LOG_FMT_FILE)
    datefmt = kwargs.get("datefmt", LOG_DATEFMT_FILE)
    htype = logging.handlers.RotatingFileHandler

    path = get_path(obj=file_path)
    path.mkdir(mode=file_path_mode, parents=True, exist_ok=True)

    handler = add_handler(
        obj=obj,
        level=level,
        htype=htype,
        fmt=fmt,
        datefmt=datefmt,
        hname=hname,
        filename=str(path / file_name),
        maxBytes=max_mb * 1024 * 1024,
        backupCount=max_files,
    )
    handler.PATH = path
    return handler


def add_null(obj, traverse=True, **kwargs):
    """Add a NullHandler to a logger if it has no handlers.

    Args:
        obj (:obj:`logging.Logger`): logger obj to add handler to
        traverse (:obj:`bool`, optional): default ``True`` -

            * if ``True`` traverse the logger obj supplied up to the root logger
              to see if there are any nullhandlers attached
            * if ``False`` only check the logger obj supplied
        **kwargs:
            * hname (:obj:`str`): default: ``NULL`` - name to assign to handler

    Returns:
        :obj:`logging.NullHandler`: handler that was added to logger obj
    """
    hname = kwargs.get("hname", "NULL")
    found = find_handlers(obj=obj, hname=hname, traverse=traverse)
    htype = logging.NullHandler
    if found:
        return None
    return add_handler(obj=obj, htype=htype, hname=hname, fmt="", datefmt="", level="")


def add_handler(obj, htype, level, hname, fmt, datefmt, **kwargs):
    """Add a handler to a logger obj.

    Args:
        obj (:obj:`logging.Logger`): logger obj to add handler to
        htype (:class:`logging.Handler`): handler class to instantiate
        level (:obj:`str`): level to assign to handler obj
        hname (:obj:`str`): name to assign to handler obj
        fmt (:obj:`str`): logging format to assign to handler obj
        datefmt (:obj:`str`): date format to assign to handler obj
        **kwargs: passed to instantiation of htype

    Returns:
        :obj:`logging.Handler`: handler that was created and added
    """
    handler = htype(**kwargs)

    if hname:
        handler.name = hname

    if fmt:
        handler.setFormatter(logging.Formatter(fmt=fmt, datefmt=datefmt))

    if level:
        set_log_level(obj=handler, level=level)

    obj.addHandler(handler)
    return handler


def del_stderr(obj, traverse=True, **kwargs):
    """Remove the STDERR StreamHandler from a logger if found.

    Args:
        obj (:obj:`logging.Logger`): logger obj to remove handler from
        traverse (:obj:`bool`, optional): default ``True`` -

            * if ``True`` traverse the logger obj supplied up to the root logger
              to see if there are any attached handlers named with hname
            * if ``False`` only check the logger obj supplied for handlers named with
              hname
        **kwargs:
            * hname (:obj:`str`):
              default: :data:`axonius_api_client.LOG_NAME_FILE` -
              name of handler to search for and remove

    Returns:
        :obj:`logging.Handler`:
            dict handler handler name->[handler objects] mapping of found and
            removed handlers
    """
    hname = kwargs.get("hname", LOG_NAME_STDERR)
    htype = logging.StreamHandler
    return del_handler(obj=obj, hname=hname, htype=htype, traverse=traverse)


def del_stdout(obj, traverse=True, **kwargs):
    """Remove the STDOUT StreamHandler from a logger if found.

    Args:
        obj (:obj:`logging.Logger`): logger obj to remove handler from
        traverse (:obj:`bool`, optional): default ``True`` -

            * if ``True`` traverse the logger obj supplied up to the root logger
              to see if there are any attached handlers named with hname
            * if ``False`` only check the logger obj supplied for handlers named with
              hname
        **kwargs:
            * hname (:obj:`str`):
              default: :data:`axonius_api_client.LOG_NAME_STDOUT` -
              name of handler to search for and remove

    Returns:
        :obj:`logging.Handler`:
            dict handler handler name->[handler objects] mapping of found and
            removed handlers
    """
    hname = kwargs.get("hname", LOG_NAME_STDOUT)
    htype = logging.StreamHandler
    return del_handler(obj=obj, hname=hname, htype=htype, traverse=traverse)


def del_file(obj, traverse=True, **kwargs):
    """Remove the RotatingFileHandler from a logger if found.

    Args:
        obj (:obj:`logging.Logger`): logger obj to remove handler from
        traverse (:obj:`bool`, optional): default ``True`` -

            * if ``True`` traverse the logger obj supplied up to the root logger
              to see if there are any attached handlers named with hname
            * if ``False`` only check the logger obj supplied for handlers named with
              hname
        **kwargs:
            * hname (:obj:`str`):
              default: :data:`axonius_api_client.LOG_NAME_FILE` -
              name of handler to search for and remove

    Returns:
        :obj:`logging.Handler`:
            dict handler handler name->[handler objects] mapping of found and
            removed handlers
    """
    hname = kwargs.get("hname", LOG_NAME_FILE)
    htype = logging.handlers.RotatingFileHandler
    return del_handler(obj=obj, hname=hname, htype=htype, traverse=traverse)


def del_null(obj, traverse=True, **kwargs):
    """Remove the NullHandler from a logger if found.

    Args:
        obj (:obj:`logging.Logger`): logger obj to remove handler from
        traverse (:obj:`bool`, optional): default ``True`` -

            * if ``True`` traverse the logger obj supplied up to the root logger
              to see if there are any attached handlers named with hname
            * if ``False`` only check the logger obj supplied for handlers named with
              hname
        **kwargs:
            * hname (:obj:`str`): default: NULL -
              name of handler to search for and remove

    Returns:
        :obj:`logging.Handler`:
            dict handler handler name->[handler objects] mapping of found and
            removed handlers
    """
    hname = kwargs.get("hname", "NULL")
    htype = logging.NullHandler
    return del_handler(obj=obj, hname=hname, htype=htype, traverse=traverse)


def del_handler(obj, hname="", htype=None, traverse=True):
    """Remove the NullHandler from a logger if found.

    Args:
        obj (:obj:`logging.Logger`): logger obj to remove handler from
        hname (:obj:`str`, optional): default ``""`` -
            name of handler to find and remove
        htype (:class:`object`, optional): default ``None`` -
            type of handler to find and remove
        traverse (:obj:`bool`, optional): default ``True`` -

            * if ``True`` traverse the logger obj supplied up to the root logger
              to see if there are any attached handlers named with hname
            * if ``False`` only check the logger obj supplied for handlers named with
              hname

    Returns:
            dict handler handler name->[handler objects] mapping of found and
            removed handlers
    """
    found = find_handlers(obj=obj, hname=hname, htype=htype, traverse=traverse)
    for name, handlers in found.items():
        for handler in handlers:
            logging.getLogger(name).removeHandler(handler)
    return found


def find_handlers(obj, hname="", htype=None, traverse=True):
    """Remove the NullHandler from a logger if found.

    Notes:
        * will remove handler if hname supplied and handler obj name matches
        * will remove handler if htype supplied and handler obj type matches

    Args:
        obj (:obj:`logging.Logger`): logger obj to remove handler from
        hname (:obj:`str`, optional): default ``""`` -
            name of handler to find and remove
        htype (:class:`object`, optional): default ``None`` -
            type of handler to find and remove
        traverse (:obj:`bool`, optional): default ``True`` -

            * if ``True`` traverse the logger obj supplied up to the root logger
              to see if there are any matching attached handlers
            * if ``False`` only check the logger obj supplied for any matching
              attached handlers

    Returns:
            dict handler handler name->[handler objects] mapping of found handlers
    """
    handlers = {}

    for handler in obj.handlers:
        match_name = hname and handler.name == hname
        match_type = htype and isinstance(handler, htype)

        if match_name or match_type:
            handlers[obj.name] = handlers.get(obj.name, [])

            if handler not in handlers[obj.name]:
                handlers[obj.name].append(handler)

    if obj.parent and traverse:
        found = find_handlers(
            obj=obj.parent, hname=hname, htype=htype, traverse=traverse
        )
        handlers.update(found)

    return handlers


LOG = logging.getLogger(PACKAGE_ROOT)
""":obj:`logging.Logger`: root logger used by entire package, named after package."""

add_null(obj=LOG)
gmtime()



# -*- coding: utf-8 -*-
"""API models package."""
from ...exceptions import ApiError
from . import base, base_csv, base_json, base_table
from .base import Base
from .base_csv import Csv
from .base_json import Json
from .base_table import Table

__all__ = (
    "Base",
    "Csv",
    "Json",
    "Table",
    "base",
    "base_csv",
    "base_json",
    "base_table",
)


DEFAULT_CALLBACKS_CLS = "base"
CALLBACKS_MAP = {"json": Json, "csv": Csv, "table": Table, "base": Base}


def get_callbacks_cls(export=None):
    """Get a callback class."""
    export = export or DEFAULT_CALLBACKS_CLS
    if export in CALLBACKS_MAP:
        return CALLBACKS_MAP[export]
    raise ApiError(f"Invalid export {export!r}, valids: {list(CALLBACKS_MAP)}")



# -*- coding: utf-8 -*-
"""Test suite for assets."""
import pytest
import requests
from axonius_api_client.api import mixins
from axonius_api_client.constants import MAX_PAGE_SIZE
from axonius_api_client.exceptions import (JsonError, JsonInvalid,
                                           NotFoundError, ResponseNotOk)
from axonius_api_client.tools import listify

from ...meta import QUERIES


def load_test_data(apiobj):
    """Pass."""
    apiobj.TEST_DATA = getattr(apiobj, "TEST_DATA", {})

    if not apiobj.TEST_DATA.get("fields_map"):
        apiobj.TEST_DATA["fields_map"] = fields_map = apiobj.fields.get()

    if not apiobj.TEST_DATA.get("assets"):
        apiobj.TEST_DATA["assets"] = apiobj.get(max_rows=4000, fields_map=fields_map)

    if not apiobj.TEST_DATA.get("saved_queries"):
        apiobj.TEST_DATA["saved_queries"] = apiobj.saved_query.get()

    return apiobj


class ModelMixinsBase:
    """Pass."""

    def test_model_json(self, apiobj):
        """Test that JSON is returned when is_json=True."""
        response = apiobj.request(
            path=apiobj.router.fields,
            method="get",
            raw=False,
            is_json=True,
            error_status=True,
        )
        assert isinstance(response, dict)

    def test_model_raw(self, apiobj):
        """Test that response is returned when raw=True."""
        response = apiobj.request(
            path=apiobj.router.fields,
            method="get",
            raw=True,
            is_json=True,
            error_status=True,
        )
        assert isinstance(response, requests.Response)

    def test_model_text(self, apiobj):
        """Test that str is returned when raw=False and is_json=False."""
        response = apiobj.request(
            path=apiobj.router.fields,
            method="get",
            raw=False,
            is_json=False,
            error_status=True,
        )
        assert isinstance(response, str)

    def test_model_json_error(self, apiobj):
        """Test exc thrown when json has error status."""
        with pytest.raises(JsonError):
            apiobj.request(
                path=apiobj.router.root + "/badwolf", method="get", error_status=False
            )

    def test_model_no_json_error(self, apiobj):
        """Test exc thrown when status code != 200."""
        with pytest.raises(ResponseNotOk):
            apiobj.request(
                path=apiobj.router.root + "/badwolf",
                method="get",
                error_status=True,
                is_json=False,
            )

    def test_model_json_invalid(self, apiobj):
        """Test exc thrown when invalid json."""
        with pytest.raises(JsonInvalid):
            apiobj.request(path="", method="get")

    def test_model_json_invalid_text(self, apiobj):
        """Test that str is returned when is_json=True and error_json_invalid=False."""
        response = apiobj.request(path="", method="get", error_json_invalid=False)
        assert isinstance(response, str)

    def test_model_child(self, apiobj):
        """Pass."""
        child = mixins.ChildMixins(parent=apiobj)
        assert str(apiobj) in str(child)
        assert repr(apiobj) in repr(child)


class AssetsPrivate:
    """Pass."""

    def test_private_get_page_stop_rows_stop(self, apiobj):
        """Pass."""
        state = {
            "max_rows": None,
            "rows_processed": 1,
            "stop": True,
            "stop_msg": "fakestop",
            "max_pages": None,
            "page_num": 1,
        }
        result = apiobj._get_page_stop_rows(row={}, rows=[{}], state=state, store={})
        assert result is True

    def test_private_get_page_stop_fetch_stop(self, apiobj):
        """Pass."""
        state = {
            "max_rows": None,
            "rows_processed": 1,
            "stop": True,
            "stop_msg": "fakestop",
            "max_pages": None,
            "page_num": 1,
        }
        result = apiobj._get_page_stop_fetch(rows=[{}], state=state, store={})
        assert result is True

    def test_private_get(self, apiobj):
        """Pass."""
        data = apiobj._get(page_size=1)
        assert isinstance(data, dict)
        assert isinstance(data["assets"], list)
        assert len(data["assets"]) == 1

    def test_private_get_normal_query(self, apiobj):
        """Pass."""
        data = apiobj._get(query=QUERIES["not_last_seen_day"], page_size=1)
        assert isinstance(data, dict)
        assert isinstance(data["assets"], list)
        assert len(data["assets"]) == 1

    def test_private_get_cursor(self, apiobj):
        """Pass."""
        data = apiobj._get_cursor(page_size=1)
        assert isinstance(data, dict)
        assert isinstance(data["assets"], list)
        assert len(data["assets"]) == 1

    def test_private_get_cursor_query(self, apiobj):
        """Pass."""
        query = QUERIES["not_last_seen_day"]
        data = apiobj._get_cursor(query=query, page_size=1)
        assert isinstance(data, dict)
        assert isinstance(data["assets"], list)
        assert len(data["assets"]) == 1

    def test_private_get_by_id(self, apiobj):
        """Pass."""
        asset = apiobj.TEST_DATA["assets"][0]
        id = asset["internal_axon_id"]
        data = apiobj._get_by_id(id=id)
        assert isinstance(data, dict)
        assert data["internal_axon_id"] == id

    def test_private_count(self, apiobj):
        """Pass."""
        data = apiobj._count()
        assert isinstance(data, int)

    def test_private_count_query(self, apiobj):
        """Pass."""
        query = QUERIES["not_last_seen_day"]
        data = apiobj._count(query=query)
        assert isinstance(data, int)

    def test_private_build_query(self, apiobj):
        """Pass."""
        pre_query = QUERIES["not_last_seen_day"]
        post_query = QUERIES["not_last_seen_day"]

        inner = 'hostname == "badwolf"'
        pre = f"{pre_query} and"
        post = f"and {post_query}"

        query = apiobj._build_query(inner=inner)
        assert query == f"({inner})"

        query = apiobj._build_query(inner=inner, not_flag=True)
        assert query == f"(not ({inner}))"

        query = apiobj._build_query(inner=inner, pre=pre)
        assert query == f"{pre} ({inner})"

        query = apiobj._build_query(inner=inner, post=post)
        assert query == f"({inner}) {post}"

        query = apiobj._build_query(inner=inner, pre=pre, post=post)
        assert query == f"{pre} ({inner}) {post}"

        query = apiobj._build_query(inner=inner, pre=pre, post=post, not_flag=True)
        assert query == f"{pre} (not ({inner})) {post}"


class AssetsPublic:
    """Pass."""

    def test_count(self, apiobj):
        """Pass."""
        data = apiobj.count()
        assert isinstance(data, int)

    def test_count_query(self, apiobj):
        """Pass."""
        query = QUERIES["not_last_seen_day"]
        data = apiobj.count(query=query)
        assert isinstance(data, int)

    def test_count_by_saved_query(self, apiobj):
        """Pass."""
        sq = apiobj.TEST_DATA["saved_queries"][0]
        sq_name = sq["name"]
        data = apiobj.count_by_saved_query(name=sq_name)
        assert isinstance(data, int)

    def test_get_cursor_no_generator_no(self, apiobj):
        """Pass."""
        rows = apiobj.get(use_cursor=False, generator=False, max_rows=1)

        assert not rows.__class__.__name__ == "generator"
        assert "cached" not in apiobj.auth.http.LAST_RESPONSE.request.url
        check_assets(rows)
        assert len(rows) == 1

    def test_get_cursor_no_generator_yes(self, apiobj):
        """Pass."""
        gen = apiobj.get(use_cursor=False, generator=True, max_rows=1)

        assert not isinstance(gen, list)
        assert gen.__class__.__name__ == "generator"

        rows = [x for x in gen]

        assert "cached" not in apiobj.auth.http.LAST_RESPONSE.request.url
        check_assets(rows)
        assert len(rows) == 1

    def test_get_cursor_yes_generator_no(self, apiobj):
        """Pass."""
        if getattr(apiobj, "NO_CURSOR_OVERRIDE", False):
            pytest.skip(f"{apiobj} has no cursor override")

        rows = apiobj.get(use_cursor=True, generator=False, max_rows=1)

        assert not rows.__class__.__name__ == "generator"
        assert "cached" in apiobj.auth.http.LAST_RESPONSE.request.url
        check_assets(rows)
        assert len(rows) == 1

    def test_get_cursor_yes_generator_yes(self, apiobj):
        """Pass."""
        if getattr(apiobj, "NO_CURSOR_OVERRIDE", False):
            pytest.skip(f"{apiobj} has no cursor override")

        gen = apiobj.get(use_cursor=True, generator=True, max_rows=1)

        assert gen.__class__.__name__ == "generator"

        rows = [x for x in gen]
        assert "cached" in apiobj.auth.http.LAST_RESPONSE.request.url
        check_assets(rows)
        assert len(rows) == 1

    def test_get_page_size_over_max(self, apiobj):
        """Pass."""
        rows = apiobj.get(page_size=3000, max_pages=1)
        check_assets(rows)
        assert len(rows) <= MAX_PAGE_SIZE

    def test_get_maxpages(self, apiobj):
        """Pass."""
        rows = apiobj.get(page_size=20, max_pages=1)
        check_assets(rows)
        assert len(rows) == 20

    def test_get_id(self, apiobj):
        """Pass."""
        asset = apiobj.TEST_DATA["assets"][0]
        id = asset["internal_axon_id"]

        row = apiobj.get_by_id(id=id)
        check_asset(row)
        assert row["internal_axon_id"] == id

    def test_get_id_error(self, apiobj):
        """Pass."""
        with pytest.raises(NotFoundError):
            apiobj.get_by_id(id="badwolf")

    def test_get_by_saved_query(self, apiobj):
        """Pass."""
        sq = apiobj.TEST_DATA["saved_queries"][0]
        sq_name = sq["name"]
        sq_fields = sq["view"]["fields"]

        rows = apiobj.get_by_saved_query(
            name=sq_name, max_rows=1, fields_map=apiobj.TEST_DATA["fields_map"],
        )
        check_assets(rows)

        last_fields = apiobj._LAST_GET["fields"].split(",")
        assert sq_fields == last_fields

    def test_get_by_value(self, apiobj):
        """Pass."""
        field = apiobj.TEST_DATA["field_main"]

        values = get_field_values(rows=apiobj.TEST_DATA["assets"], field=field)
        value = values[0]

        rows = apiobj.get_by_value(
            value=value, field=field, fields_map=apiobj.TEST_DATA["fields_map"],
        )
        check_assets(rows)
        assert len(rows) == 1

        rows_values = get_field_values(rows=rows, field=field)
        assert value in rows_values

    def test_get_by_value_not(self, apiobj):
        """Pass."""
        field = apiobj.TEST_DATA["field_main"]

        values = get_field_values(rows=apiobj.TEST_DATA["assets"], field=field)
        value = values[0]

        rows = apiobj.get_by_value(
            value=value,
            field=field,
            not_flag=True,
            fields_map=apiobj.TEST_DATA["fields_map"],
        )
        check_assets(rows)

        rows_values = get_field_values(rows=rows, field=field)
        assert value not in rows_values

    def test_get_by_values(self, apiobj):
        """Pass."""
        field = apiobj.TEST_DATA["field_main"]

        values = get_field_values(rows=apiobj.TEST_DATA["assets"], field=field)
        values = values[0:2]

        rows = apiobj.get_by_values(
            values=values, field=field, fields_map=apiobj.TEST_DATA["fields_map"],
        )
        check_assets(rows)

        rows_values = get_field_values(rows=rows, field=field)
        for value in values:
            assert value in rows_values

    def test_get_by_values_not(self, apiobj):
        """Pass."""
        field = apiobj.TEST_DATA["field_main"]

        values = get_field_values(rows=apiobj.TEST_DATA["assets"], field=field)
        values = values[0:2]

        rows = apiobj.get_by_values(
            values=values,
            field=field,
            not_flag=True,
            fields_map=apiobj.TEST_DATA["fields_map"],
        )
        check_assets(rows)

        rows_values = get_field_values(rows=rows, field=field)
        for value in values:
            assert value not in rows_values

    def test_get_by_value_regex(self, apiobj):
        """Pass."""
        field = apiobj.TEST_DATA["field_main"]

        values = get_field_values(rows=apiobj.TEST_DATA["assets"], field=field)
        value = values[0]
        regex_value = value[0:5]

        rows = apiobj.get_by_value_regex(
            value=regex_value, field=field, fields_map=apiobj.TEST_DATA["fields_map"],
        )
        check_assets(rows)

        rows_values = get_field_values(rows=rows, field=field)
        assert value in rows_values

    def test_get_by_value_regex_not(self, apiobj):
        """Pass."""
        field = apiobj.TEST_DATA["field_main"]

        values = get_field_values(rows=apiobj.TEST_DATA["assets"], field=field)
        value = values[0]
        regex_value = value[0:5]

        rows = apiobj.get_by_value_regex(
            value=regex_value,
            field=field,
            not_flag=True,
            fields_map=apiobj.TEST_DATA["fields_map"],
        )
        check_assets(rows)

        rows_values = get_field_values(rows=rows, field=field)
        assert value not in rows_values


def get_field_values(rows, field):
    """Pass."""
    values = [x[field] for x in rows if x.get(field)]
    values = [x for y in values for x in listify(y)]
    return values


def check_assets(rows):
    """Pass."""
    assert isinstance(rows, list)
    for row in rows:
        check_asset(row)


def check_asset(row):
    """Pass."""
    assert isinstance(row, dict)
    assert row["internal_axon_id"]



# -*- coding: utf-8 -*-
"""API models package."""
from . import (adapters, assets, enforcements, entry, mixins, parsers, routers,
               system)
from .adapters import Adapters
from .assets import Devices, Users
from .enforcements import Enforcements, RunAction
from .entry import Entry
from .system import System

__all__ = (
    "Users",
    "Devices",
    "Adapters",
    "Enforcements",
    "RunAction",
    "System",
    "Entry",
    "routers",
    "assets",
    "adapters",
    "enforcements",
    "mixins",
    "system",
    "parsers",
    "entry",
)
